#include <fstream>
#include <iostream>

#include <arc/notify.h>
#include <arc/ftpcontrol.h>
#include <arc/stringconv.h>
#include <arc/common.h>

#include "globus_ftp_control.h"

FTPControl::FTPControl() throw(FTPControlError) {
	if (globus_ftp_control_handle_init(&control_handle) != GLOBUS_SUCCESS) {
		throw FTPControlError("Failed to initialize globus ftp-control "
		                      "handle.");
	}

	isconnected = false;
}


FTPControl::~FTPControl() {
	globus_ftp_control_handle_destroy(&control_handle);
}


void FTPControl::Upload(const std::string& localfile,
                        const URL& url,
                        bool disconnectafteruse,
                        int timeout) throw(FTPControlError) {

	if (url.Protocol()!="gsiftp")
		throw FTPControlError("Bad url passed to FTPControl");

	Connect(url, timeout);

	int fd = open(localfile.c_str(), O_RDONLY);
	if (fd==-1)
		throw FTPControlError("File " + localfile + " does not exist");

	notify(DEBUG) << "Opened file " << localfile << " for reading."
	              << std::endl;

	SetupReadWriteOperation(timeout);
	std::string resp = SendCommand("STOR " + url.Path(), timeout);

	if (globus_ftp_control_data_connect_write(&control_handle,
	                                          DataConnectCallback,
	                                          (void*)this)!=GLOBUS_SUCCESS) {
		throw FTPControlError("Failed to create data-connection for the "
		                      "writing.");
	}

	WaitForCallback(timeout);

	notify(DEBUG) << "Uploading file " << localfile << std::endl;

	const unsigned int maxsize = 65535;
	char filebuffer[maxsize];

	bool eof = GLOBUS_FALSE;
	unsigned long long offset = 0;
	int len = 0;

	do {
		len = read(fd, filebuffer, maxsize);
		if (len==-1) throw FTPControlError("Error reading " + localfile);
		notify(VERBOSE) << "Writing filebuffer of size " << len << std::endl;

		if (len==0) eof = GLOBUS_TRUE;
		if (globus_ftp_control_data_write(&control_handle,
		                                  (globus_byte_t*)filebuffer,
		                                  len,
										  offset,
		                                  eof,
		                                  &DataReadWriteCallback,
		                                  this)!=GLOBUS_SUCCESS) {
			throw FTPControlError("Failed writing data to data-connection.");
		}

		WaitForCallback(timeout);
		offset += len;
	} while (len!=0);

	close(fd);

	/* Wait for final callback. Apparantly this is called when data-writing
	   has finished. */
	WaitForCallback(timeout);

	if (disconnectafteruse) Disconnect(url, timeout);

	notify(INFO) << "File " << localfile << " uploaded." << std::endl;
}


void FTPControl::Download(const URL& url,
                          const std::string& localfile,
                          bool disconnectafteruse,
                          int timeout) throw(FTPControlError) {

	if (url.Protocol()!="gsiftp")
		throw FTPControlError("Bad url passed to FTPControl");

	std::string file(localfile);
	if (file.empty()) {
		std::string::size_type p = url.str().rfind("/");
		file = url.str().substr(p+1);
	}

	int fd = open(file.c_str(),
	              O_RDWR | O_CREAT | O_EXCL,
	              S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd==-1)
		throw FTPControlError("File " + file + " could not be created.");

	notify(DEBUG) << "Opened file " << file << " for writing." << std::endl;

	Connect(url, timeout);
	SetupReadWriteOperation(timeout);
	std::string resp = SendCommand("RETR " + url.Path(), timeout);

	if (globus_ftp_control_data_connect_read(&control_handle,
	                                         DataConnectCallback,
	                                         (void*)this)!=GLOBUS_SUCCESS) {
		throw FTPControlError("Failed to create data-connection for the "
		                      "reading.");
	}

	WaitForCallback(timeout);

	notify(DEBUG) << "Downloading file " << url.str() << std::endl;

	const unsigned int maxsize = 65535;
	char filebuffer[maxsize];

	eof = GLOBUS_FALSE;

	do {
		if (globus_ftp_control_data_read(&control_handle,
		                                 (globus_byte_t*)filebuffer,
		                                 maxsize,
		                                 &DataReadWriteCallback,
		                                 this)!=GLOBUS_SUCCESS) {
			throw FTPControlError("Failed writing data to data-connection.");
		}

		WaitForCallback(timeout);
		int len = write(fd, filebuffer, buffer_length);
		notify(VERBOSE) << "Wrote buffer of length " << buffer_length
		                << std::endl;
		if (len==-1) {
			close(fd);
			throw FTPControlError("Error writing local-file during download");
		}

	} while (eof!=GLOBUS_TRUE);

	close(fd);

	/* Final callback called at the end of the read-operation. */
	WaitForCallback(timeout);
	if (disconnectafteruse) Disconnect(url, timeout);

	notify(INFO) << "File " << url.str() << " downloaded." << std::endl;
}


std::list<FileInfo> FTPControl::ListDir(const URL& url,
                                        bool disconnectafteruse,
                                        int timeout) throw(FTPControlError) {

	if (url.Protocol()!="gsiftp")
		throw FTPControlError("Bad url passed to FTPControl");

	Connect(url, timeout);
	if (url.Path().size()>0)
		SendCommand("CWD " + url.Path(), timeout);

	SetupReadWriteOperation(timeout);
	SendCommand("MLSD", timeout);
	if (globus_ftp_control_data_connect_read(&control_handle,
	                                         DataConnectCallback,
	                                         (void*)this)!=GLOBUS_SUCCESS) {
		throw FTPControlError("Failed to create data-connection for the "
		                      "reading.");
	}

	WaitForCallback(timeout);

	std::string dirlisting;

	const unsigned int maxsize = 65536;
	char filebuffer[maxsize+1];
	filebuffer[maxsize] = 0;
	eof = GLOBUS_FALSE;

	do {
		if (globus_ftp_control_data_read(&control_handle,
		                                 (globus_byte_t*)filebuffer,
		                                 maxsize,
		                                 &DataReadWriteCallback,
		                                 this)!=GLOBUS_SUCCESS) {
			throw FTPControlError("Failed writing data to data-connection.");
		}

		WaitForCallback(timeout);
		if (buffer_length<maxsize) filebuffer[buffer_length] = 0;
		dirlisting.append(filebuffer);
	} while (eof!=GLOBUS_TRUE);

	/* Final callback called at the end of the read-operation. */
	WaitForCallback(timeout);

	notify(VERBOSE) << "dirlisting: " << std::endl << dirlisting << std::endl;

	std::list<FileInfo> dirlist;
	std::string::size_type pos = 0;
	std::string::size_type pos2 = 0;

	while ( (pos = dirlisting.find("\r\n", pos2))!=std::string::npos) {
		std::string fileinfo = dirlisting.substr(pos2, pos-pos2);
		
		FileInfo info;
		std::string::size_type p = fileinfo.find(" ");
		info.filename = url.Path() + "/" + fileinfo.substr(p+1);

		info.isdir = false;
		p = fileinfo.find("type=");
		if (fileinfo.substr(p+5, 3)=="dir") info.isdir = true;

		p = fileinfo.find("size=");
		std::string::size_type q = fileinfo.find(";", p+1);
		info.size = stringtoull(fileinfo.substr(p+5, q-p-5));	

		pos2 = pos + 2;

		dirlist.push_back(info);
	}

	if (disconnectafteruse) Disconnect(url, timeout);
	return dirlist;
}


std::list<FileInfo> FTPControl::RecursiveListDir(const URL& url,
                                                 bool disconnectafteruse,
                                                 int timeout)
throw(FTPControlError) {

	if (url.Protocol()!="gsiftp")
		throw FTPControlError("Bad url passed to FTPControl");

	Connect(url, timeout);
	std::list<FileInfo> files = ListDir(url, false, timeout);

	std::string ustr = url.Protocol() + "://" + url.Host();

	std::list<FileInfo>::iterator it = files.begin();
	while (it != files.end()) {
		if (it->isdir) {
			URL newurl(ustr + it->filename);
			std::list<FileInfo> more = ListDir(newurl, false, timeout);
			files.insert(files.end(), more.begin(), more.end());
		}
		it++;
	}

	if (disconnectafteruse) Disconnect(url, timeout);
	return files;
}


void FTPControl::DownloadDirectory(const URL& url,
                                   bool disconnectafteruse,
                                   int timeout) throw(FTPControlError) {

	mode_t dirmode = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
	
	// create main directory
	std::string path = url.Path();
	std::string::size_type pos = path.rfind('/');
	if (pos!=std::string::npos) {
		std::string maindir = path.substr(pos+1);
		if (!maindir.empty()) mkdir(maindir.c_str(), dirmode);
	}

	std::list<FileInfo> allfiles = RecursiveListDir(url,
	                                                false,
	                                                timeout);


	std::list<FileInfo>::iterator it;
	for (it = allfiles.begin(); it != allfiles.end(); it++) {
		if (!it->isdir) continue;
		std::string path = it->filename;
		path = path.substr(1);

		int err = mkdir(path.c_str(), dirmode);
		if (err==-1)
			throw FTPControlError("Could not create the necessary directory "
			                      "structure for downloading the files.");
	}

	std::string newstr = url.Protocol() + "://" + url.Host() + "/";

	for (it = allfiles.begin(); it != allfiles.end(); it++) {
		if (it->isdir) continue;

		std::string filename = it->filename;
		filename = filename.substr(1);
		URL newurl(newstr + filename);
		Download(newurl, filename, false);
	}

	if (disconnectafteruse) Disconnect(url, timeout);
}


unsigned long long FTPControl::Size(const URL& url,
                                    bool disconnectafteruse,
                                    int timeout) throw(FTPControlError) {

	if (url.Protocol()!="gsiftp")
		throw FTPControlError("Bad url passed to FTPControl");

	Connect(url, timeout);
	std::string resp = SendCommand("SIZE " + url.Path());
	if (disconnectafteruse) Disconnect(url, timeout);

	notify(DEBUG) << resp << std::endl;
	if (resp.size()==0) throw FTPControlError("Server returned nothing.");

	resp = resp.substr(17);
	return stringtoull(resp);
}


void FTPControl::Connect(const URL& url,
                         int timeout) throw(FTPControlError) {

	if (isconnected) return;

	notify(DEBUG) << "Connecting to server " << url.Host() << std::endl;
	if (globus_ftp_control_connect(&control_handle,
	                               (char*)url.Host().c_str(),
	                               url.Port(),
	                               &FTPControlCallback,
	                               this) != GLOBUS_SUCCESS) {
		throw FTPControlError("Failed to connect to server");
	}

	WaitForCallback(timeout);
	isconnected = true;

	notify(DEBUG) << "Authenticating to server " << url.Host() << std::endl;
	globus_ftp_control_auth_info_t auth;
	globus_ftp_control_auth_info_init(&auth,
	                                  GSS_C_NO_CREDENTIAL,
	                                  GLOBUS_TRUE,
	                                  "ftp",
	                                  "user@",
	                                  GLOBUS_NULL,
	                                  GLOBUS_NULL);

	// authenticate to server
	if (globus_ftp_control_authenticate(&control_handle,
	                                    &auth,
	                                    GLOBUS_TRUE,
	                                    FTPControlCallback,
	                                    this) != GLOBUS_SUCCESS) {
		Disconnect(url, timeout);
		throw FTPControlError("Failed authenticate to server: Call to "
		                      "globus_ftp_control_authenticate failed.");
	}

	try {
		WaitForCallback(timeout);
	} catch(FTPControlError e) {
		std::string errstr = errorstring;
		Disconnect(url, timeout);
		throw FTPControlError("Failed to authenticate to server: " + errstr);
	}

	notify(INFO) << "Connection to " << url.Host() << " established."
	             << std::endl;
}


void FTPControl::Disconnect(const URL& url,
                            int timeout) throw(FTPControlError) {

	if (!isconnected) return;

	notify(DEBUG) << "Closing connection to " << url.Host() << std::endl;

	bool closed = true; // at least let's hope that.
	if (globus_ftp_control_quit(&control_handle,
	                            &FTPControlCallback,
	                            this)==GLOBUS_SUCCESS) {
		try {
			WaitForCallback(timeout);
		} catch(FTPControlError e) {
			closed = false;
		}
	}

	if (closed==false) {
		// Ordinary close failed. Let's force close.
		notify(DEBUG) << "Forcing close connection to " << url.Host()
		              << std::endl;
		if (globus_ftp_control_force_close(&control_handle,
		                                   &FTPControlCallback,
		                                   this)!=GLOBUS_SUCCESS) {
			throw FTPControlError("Failed calling "
			                      "globus_ftp_control_force_close.");
		}

		WaitForCallback(timeout);
	}

	isconnected = false;
	notify(INFO) << "Connection to " << url.Host() << " closed." << std::endl;
}


std::string FTPControl::SendCommand(const std::string& command,
                                    int timeout) throw(FTPControlError) {

	std::string newcommand(command);

	if (newcommand.length() > 0) {
		notify(DEBUG) << "Sending command: " << newcommand << std::endl;
		newcommand += "\r\n";
		if (globus_ftp_control_send_command(&control_handle,
		                                    newcommand.c_str(),
		                                    FTPControlCallback,
		                                    this) != GLOBUS_SUCCESS) {
			throw FTPControlError("Sending command "+ newcommand +" failed.");
		}
	}

	WaitForCallback(timeout);

	if (server_resp.size()>0)
		notify(VERBOSE) << server_resp << std::endl;

	return server_resp;
}


void FTPControl::SetupReadWriteOperation(int timeout) throw(FTPControlError) {

	std::string resp = SendCommand("DCAU N", timeout);
	resp = SendCommand("TYPE I", timeout);

	resp = SendCommand("PASV", timeout);
	std::string::size_type pos;
	if ( (pos = resp.find('('))==std::string::npos)
		throw FTPControlError("Could not parse server response.");
	resp = resp.substr(pos+1);
	
	if ( (pos = resp.find(')'))==std::string::npos)
		throw FTPControlError("Could not parse server response.");
	resp = resp.substr(0, pos);
	
	globus_ftp_control_host_port_t passive_addr;
	passive_addr.port = 0;
	unsigned short port_low, port_high;

	if (sscanf(resp.c_str(),
			   "%i,%i,%i,%i,%hu,%hu",
			   &passive_addr.host[0],
			   &passive_addr.host[1],
			   &passive_addr.host[2],
			   &passive_addr.host[3],
			   &port_high,
			   &port_low)==6) {
		passive_addr.port = 256*port_high + port_low;
	}

	if (passive_addr.port == 0) {
		throw FTPControlError("Could not parse host and port in PASV "
		                      "response: " + resp);
	}

	if (globus_ftp_control_local_port(&control_handle,
	                                  &passive_addr)!=GLOBUS_SUCCESS) {
		throw FTPControlError("The received PASV host and address values "
		                      "are not acceptable: " + resp);
	}

	if (globus_ftp_control_local_type(&control_handle,
	                                  GLOBUS_FTP_CONTROL_TYPE_IMAGE,
	                                  0)!=GLOBUS_SUCCESS) {
		throw FTPControlError("Setting data type to IMAGE failed.");
	}
}


void FTPControl::WaitForCallback(int timeout) throw(FTPControlError) {

	notify(VERBOSE) << "Waiting for callback (timeout " << timeout << ")"
	                << std::endl;

	bool valid;
	if (!cond.Wait(valid, timeout)) {
		// timeout -- abort operation and return error
		notify(DEBUG) << "Timeout: Aborting operation" << std::endl;
		AbortOperation();
		valid = false;
	}

	if (valid==false) {
		if (!errorstring.empty()) throw FTPControlError(errorstring);
		if (!server_resp.empty()) throw FTPControlError(server_resp);
		throw FTPControlError("Unknown error.");
	}
}


void FTPControl::AbortOperation() {

	notify(VERBOSE) << "AbortOperation called." << std::endl;
	if (globus_ftp_control_abort(&control_handle,
	                             FTPControlCallback,
	                             this)!=GLOBUS_SUCCESS) {
		/* if the operation cannot be aborted here, probably the operation
		   already finished (race-condition?). We return an error
		   assuming no more callbacks will be called. */
		errorstring = "Aborting operation failed.";
		return;
	}

	/* Wait until AbortCallback is called (no timeout).
	   Pray it will be called! */
	WaitForCallback(-1);
}


void FTPControl::DataReadWriteCallback(void* arg,
                                       globus_ftp_control_handle_t* handle,
                                       globus_object_t* error,
                                       globus_byte_t* buffer,
                                       globus_size_t length,
                                       globus_off_t offset,
                                       globus_bool_t eof) {

	FTPControl* it = (FTPControl*)arg;

	if (eof==GLOBUS_TRUE) it->eof = GLOBUS_TRUE;
	it->buffer_length = length;

	FTPControlCallback(arg, handle, error, NULL);
}


void FTPControl::DataConnectCallback(void* arg,
                                     globus_ftp_control_handle_t* handle,
                                     unsigned int stripe_ndx,
                                     globus_bool_t reused,
                                     globus_object_t* error) {

	FTPControlCallback(arg, handle, error, NULL);
}


void FTPControl::FTPControlCallback(void* arg,
                                    globus_ftp_control_handle_t* handle,
                                    globus_object_t* error,
                                    globus_ftp_control_response_t* response) {

	notify(VERBOSE) << "FTPControlCallback called" << std::endl;

	FTPControl* it = (FTPControl*)arg;
	it->server_resp.resize(0);
	globus_ftp_control_response_class_t responseclass =
		GLOBUS_FTP_POSITIVE_COMPLETION_REPLY;

	std::string::size_type pos;
	if (response && (char*)response->response_buffer!=NULL) {
		it->server_resp.assign((char*)response->response_buffer,
		                       response->response_length);
		while ((pos = it->server_resp.find("\r\n"))!=std::string::npos)
			it->server_resp.erase(pos, 2);

		it->server_resp = "Server response: " + it->server_resp.substr(4);
		responseclass = response->response_class;
	}

	if (error == GLOBUS_SUCCESS) { // normal callback
		/* If response class is GLOBUS_FTP_TRANSIENT_NEGATIVE_COMPLETION_REPLY
		   or GLOBUS_FTP_PERMANENT_NEGATIVE_COMPLETION_REPLY i.e. >= 4, we
		   have an error. */
		if (responseclass>=4) {
			it->cond.Signal(false);
		} else {
			it->cond.Signal(true);
		}
	} else { // callback-error
		it->errorstring = GlobusErrorString(error);

		while ((pos = it->errorstring.find("\r\n"))!=std::string::npos)
			it->errorstring.erase(pos, 2);
		if (it->errorstring.find("end-of-file") !=std::string::npos)
			it->errorstring = "Server unexpectedly closed connection.";
		if (it->errorstring.find("GSS failure") !=std::string::npos)
			it->errorstring = "Problem with GSI credential.";

		it->cond.Signal(false);
	}
}
