#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libxml/parserInternals.h>

#include <arc/configcore.h>
#include <arc/xmlconfig.h>

#include <fstream>

void XMLConfig::fill_tree(xmlNode *node, Config& config) {

	xmlNode *cur_node = NULL;
	std::string value;
	int has_element = 0;
	std::string key;

	for (cur_node = node; cur_node; cur_node = cur_node->next) {
		if (cur_node->type == XML_ELEMENT_NODE) {
			has_element = 1;
		}
		if (cur_node->children) fill_tree(cur_node->children, config);
	}

	// No sub-elements ("attributes")
	if (has_element == 0 ) {
	for (cur_node = node; cur_node->parent->type != XML_DOCUMENT_NODE;
	     cur_node = cur_node->parent) {
		if (cur_node->type == XML_ELEMENT_NODE){

			// Adding the attribute as an element
			xmlAttr *attribute;
			for (attribute = cur_node->properties; attribute;
			     attribute = attribute->next) {
				if (strcmp((char *) attribute->name, "label") == 0) {
					key = "/" + key;
					key = (char *) attribute->children->content + key;
				}
			}

			if (key != "") key = "/" + key;

			key = (char *) cur_node->name + key;
		}
	}
	config.AddConfValue(key, (char *) node->content);
	}
}

Config XMLConfig::Read(std::istream& is) {

	Config config;
	std::string buffer,line;

	// Read stream content into buffer
	while(getline(is,line)) {
		buffer += line;
	}

	// create a parser context
	xmlParserCtxtPtr ctxt = xmlNewParserCtxt();
	if (ctxt == NULL) {
		throw ConfigError("Failed to create parser context");
	}

	// Parse and validate buffer and fill the tree
	xmlDocPtr doc;
	doc = xmlParseMemory(buffer.c_str(), buffer.length());

	// Check if parsing suceeded
	if (doc == NULL) {
		throw ConfigError("Failed xml parsing");
	} else {
	// Check if validation suceeded
		if (ctxt->valid == 0) {
		    throw ConfigError("Failed to validate xml ");
		}
	}

	// Free up the parser context
	xmlFreeParserCtxt(ctxt);

	// Walk the tree
	xmlNode *root_element = xmlDocGetRootElement(doc);

	fill_tree(root_element, config);

	xmlFreeDoc(doc);

	return config;
}

void XMLConfig::Write(Config config, std::ostream& os) {

	xmlDocPtr doc = NULL;
	xmlDtdPtr dtd = NULL;
	xmlNodePtr root_node = NULL;
	xmlNodePtr node = NULL;

	std::string root_name = "arc";

	LIBXML_TEST_VERSION;

	// Create a new document, a node and set it as a root node
	doc = xmlNewDoc(BAD_CAST "1.0");
	root_node = xmlNewNode(NULL, BAD_CAST root_name.c_str());
	xmlDocSetRootElement(doc, root_node);

	// Create DTD declaration.
	dtd = xmlCreateIntSubset(doc,
	                         BAD_CAST root_name.c_str(),
	                         NULL,
	                         BAD_CAST "nordugrid.dtd");

	// Create the content
	Options::iterator key;

	for (key = config.GetOptions().begin(); key != config.GetOptions().end();
	     key++ ) {

		// Section is everything before the last slash
		std::string section = (key->first).substr(0,(key->first).find_last_of("/"));
		// The element is defined as the text after the last slash
		std::string element = (key->first).substr( (key->first).find_last_of("/") + 1);

		std::string attribute = "";

		// Loop over values
		std::list<std::string>::iterator value;
		for (value=(key->second).begin(); value!=(key->second).end(); value++) {

			xmlXPathContextPtr context;
			std::string xpath;
			xmlXPathObjectPtr result;

			unsigned int pos = 0, pos2, level = 0;

			// Look for attributes
			while (  (pos2 = section.find("/",pos)) != std::string::npos ) {
				if (level == 0) {
					attribute = (section.substr(pos2+1)).c_str();
					// Remove the attribute part + the slash from the "section"
					section.erase(pos2);
					pos = pos2+1;
				} else {
					throw ConfigError("Many levels not supported");
				}
				level++;
			}
			xpath = "/" + root_name + "/" + section;
			if (attribute != "") {
				xpath += "[@label='" + attribute + "']";
			}

			context = xmlXPathNewContext(doc);
			result = xmlXPathEvalExpression((xmlChar*)xpath.c_str(), context);
			xmlXPathFreeContext(context);

			// Full hierarchy (no attributes) support
			if (xmlXPathNodeSetIsEmpty(result->nodesetval)){
				unsigned int pos = 0;
				unsigned int pos2;
				node = root_node;
				while ((pos2 = section.find("/",pos)) != std::string::npos) {
					node = xmlNewChild(node, NULL, BAD_CAST (section.substr(pos,pos2-pos)).c_str(), NULL);
					pos = pos2+1;
				}
				node = xmlNewChild(node, NULL, BAD_CAST (section.substr(pos,section.size()-pos)).c_str(), NULL);
			} else {
				node = (result->nodesetval)->nodeTab[0];
			}

			// Add the attribute=value as the element: <elmement label="attribute">value</element>
			xmlNewChild(node,
			            NULL,
			            BAD_CAST element.c_str(),
			            BAD_CAST value->c_str());
			if (attribute != "") {
				xmlSetProp(node,
				           BAD_CAST "label",
				           BAD_CAST attribute.c_str());
			}
		}
	}

	xmlChar *xmlbuff;
	int buffersize;
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);

	// Send the buffer to the stream
	os << xmlbuff;

	// Free the buffer
	xmlFree(xmlbuff);

	//  free the document
	xmlFreeDoc(doc);

	// Free variables allocated by parser
	xmlCleanupParser();
}
