#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <vector>
#include <map>

#ifndef UI_REMOTEQUERY
#define UI_REMOTEQUERY

class Cluster;
class Queue;


enum FileLocation {
  REMOTE,
  LOCAL,
  REMOTENOCACHE,
  LOCALNOCACHE,
  CACHED,
  NOLOCATION,
  IOERROR
};


class RemoteFileInfo {

 public:
  RemoteFileInfo (const char T, const int timeout);
  virtual ~RemoteFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const = 0;
  virtual bool IsLocal (Cluster * c) const;
  virtual int Query () = 0;
  long long int GetSize ();
  bool IsType (const char T) const;
  void RegisterCachedFile (const std::string & cluster);
  bool IsCached (const std::string & cluster) const;
 protected:
  int timeout;
  long long int size;
  bool queried;
 private:
  const char type;
  std::vector <std::string> cached;
};


#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
class RcLocationInfo {

 public:
  RcLocationInfo (const std::string & name, const std::string & uc, const std::string & path);
  ~RcLocationInfo () {};
  const std::string & GetName () const;
  const std::string & GetUrl () const;
  const std::string   GetHost () const;
 private:
  std::string name;
  std::string url;
};


class RcFileInfo : public RemoteFileInfo {

 public:
  RcFileInfo (const std::string & rc, const std::string & lfn, const int timeout);
  virtual ~RcFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual int Query ();
 private:
  const std::string rc;
  const std::string lfn;
 public:
  std::vector <RcLocationInfo> loclist;
};
#endif


#ifdef HAVE_GLOBUS_RLS_CLIENT_H
class RlsLocationInfo {

 public:
  RlsLocationInfo (const std::string & url);
  ~RlsLocationInfo () {};
  const std::string & GetUrl () const;
  const std::string   GetHost () const;
 private:
  std::string url;
};


class RlsFileInfo : public RemoteFileInfo {

 public:
  RlsFileInfo (const std::string & server, const std::string & lfn, const int timeout);
  virtual ~RlsFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual int Query ();
 private:
  const std::string server;
  const std::string lfn;
 public:
  std::vector <RlsLocationInfo> loclist;
};
#endif


class FtpFileInfo : public RemoteFileInfo {

 public:
  FtpFileInfo (const std::string & url, const int timeout);
  virtual ~FtpFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual bool IsLocal (Cluster * c) const;
  virtual int Query ();
 private:
  const std::string url;
};


class HttpFileInfo : public RemoteFileInfo {

 public:
  HttpFileInfo (const std::string & url, const int timeout);
  virtual ~HttpFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual bool IsLocal (Cluster * c) const;
  virtual int Query ();
 private:
  const std::string url;
};


class LocalFileInfo : public RemoteFileInfo {

 public:
  LocalFileInfo (const std::string & filename, const int timeout);
  virtual ~LocalFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual int Query ();
 private:
  const std::string filename;
};


class RemoteFile {

 public:
  RemoteFile (const std::string & optstring);
  virtual ~RemoteFile () {};
  virtual FileLocation Local (Cluster * c, const bool defaultcache) const;
  long long int GetSize () const;
  bool IsType (const char T) const;
  void RegisterCachedFile (const std::string & cluster) const;
 protected:
  void Register ();
  RemoteFileInfo * fileinfo;
  std::map <std::string, std::string> options;
 private:
  static std::vector <RemoteFileInfo *> filecloud;
};


#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
class RcLocation {

 public:
  RcLocation (const std::string & name, const std::string & optstring);
  ~RcLocation () {};
  const std::string & GetName () const;
  bool GetBooleanOption (const std::string & opt, bool * value) const;
 private:
  std::string name;
  std::map <std::string, std::string> options;
};


class RcFile : public RemoteFile {

 public:
  RcFile (const std::string & rc, const std::string & lfn, const std::string & locstring,
	  const std::string & optstring, const int timeout);
  virtual ~RcFile () {};
  virtual FileLocation Local (Cluster * c, const bool defaultcache) const;
 private:
  std::vector <RcLocation> loclist;
  std::map <std::string, std::string> defaultoptions;
};
#endif


#ifdef HAVE_GLOBUS_RLS_CLIENT_H
class RlsLocation {

 public:
  RlsLocation (const std::string & url, const std::string & optstring);
  ~RlsLocation () {};
  const std::string & GetUrl () const;
  bool GetBooleanOption (const std::string & opt, bool * value) const;
 private:
  std::string url;
  std::map <std::string, std::string> options;
};


class RlsFile : public RemoteFile {

 public:
  RlsFile (const std::string & server, const std::string & lfn, const std::string & locstring,
	   const std::string & optstring, const int timeout);
  virtual ~RlsFile () {};
  virtual FileLocation Local (Cluster * c, const bool defaultcache) const;
 private:
  std::vector <RlsLocation> loclist;
  std::map <std::string, std::string> defaultoptions;
};
#endif


class FtpFile : public RemoteFile {

 public:
  FtpFile (const std::string & url, const std::string & optstring, const int timeout);
  virtual ~FtpFile () {};
};


class HttpFile : public RemoteFile {

 public:
  HttpFile (const std::string & url, const std::string & optstring, const int timeout);
  virtual ~HttpFile () {};
};


class LocalFile : public RemoteFile {

 public:
  LocalFile (const std::string & filename, const std::string & optstring,
	     const int timeout);
  virtual ~LocalFile () {};
  virtual FileLocation Local (Cluster * c, const bool defaultcache) const;
};


class RemoteFileQuery {

 public:
  RemoteFileQuery () {};
  RemoteFileQuery (const RemoteFileQuery & query);
  ~RemoteFileQuery ();
  RemoteFileQuery & operator= (const RemoteFileQuery & query);
  void AddFile (RemoteFile * const file);
  int AddSizes (Cluster * c,
		long long int * cachesize,
		long long int * sessdirsize,
		long long int * remotesize,
		long long int * localsize,
		const bool defaultcache) const;
  int RegisterCachedFiles (Cluster * c, const bool defaultcache) const;
 private:
  std::vector <RemoteFile *> filelist;
};


#endif
